/*
   Copyright (C) 2011 Sérgio Martins <sergio.martins@kdab.com>
   Copyright (C) 2012 Sérgio Martins <iamsergio@gmail.com>

   This library is free software; you can redistribute it and/or modify it
   under the terms of the GNU Library General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This library is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
   License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to the
   Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA.
*/

#ifndef _AKONADI_CALENDARBASE_H_
#define _AKONADI_CALENDARBASE_H_

#include "akonadi-calendar_export.h"

#include <akonadi/item.h>
#include <akonadi/collection.h>
#include <kcalcore/memorycalendar.h>
#include <kcalcore/incidence.h>
#include <KDE/KDateTime>

namespace Akonadi {

class CalendarBasePrivate;
class IncidenceChanger;

/**
 * A calendar consisting of multiple calendars
 * 
 * TODO: The calendar should probably eventually be more than just a string.
 */
class AKONADI_CALENDAR_EXPORT MultiCalendar : public KCalCore::MemoryCalendar
{
public:
    MultiCalendar(const KTimeZone &);
    virtual bool addIncidenceToCalendar(const QString &calendar, const KCalCore::Incidence::Ptr &incidence);
    virtual bool deleteIncidenceFromCalendar(const QString &calendar, const KCalCore::Incidence::Ptr &incidence);
    virtual KCalCore::Incidence::List incidencesFromCalendar(const QString &calendar) const;
    virtual KCalCore::Incidence::Ptr incidence(const QString &calendar, const QString &uid, const KDateTime &recurrenceId = KDateTime()) const;
    virtual KCalCore::Incidence::List incidenceExceptions(const KCalCore::Incidence::Ptr &incidence) const;
    QStringList calendars(const QString &uid, const KDateTime &recurrenceId = KDateTime()) const;
    // Works only with pointers that are already in the calendar
    QString calendar(const KCalCore::Incidence::Ptr &incidence) const;
    QString uniqueInstanceIdentifier(const KCalCore::Incidence::Ptr &incidence) const;

    //TODO the whole parent handling belongs in a base class and not here
    KCalCore::Incidence::List childIncidences(const KCalCore::Incidence::Ptr &) const;

    virtual void incidenceUpdated(KCalCore::IncidenceBase *);

protected:
    virtual QString realUid(const KCalCore::Incidence::Ptr &incidence) const;
    virtual QString realIdentifier(const KCalCore::Incidence::Ptr &incidence) const;

private:
    void addParentInformation(const KCalCore::Incidence::Ptr &inc, const KCalCore::Incidence::Ptr &parentIncidence);
    void cleanupParentInformation(const KCalCore::Incidence::Ptr &inc);

    QMultiHash<QString, KCalCore::Incidence::Ptr> mIncidenceByCalendar;
    QHash<KCalCore::Incidence::Ptr, QString> mCalendarByIncidence;
    QHash<KCalCore::Incidence::Ptr, KCalCore::Incidence::List> mChildrenByIncidence;
    QHash<KCalCore::Incidence::Ptr, KCalCore::Incidence::Ptr> mParentByIncidence;
};


/**
* @short The base class for all akonadi aware calendars.
*
* Because it inherits KCalCore::Calendar, it provides seamless integration
* with KCalCore and KCalUtils libraries eliminating any need for adapter
* ( akonadi<->KCalCore ) classes.
*
* @see ETMCalendar
* @see FetchJobCalendar
*
* @author Sérgio Martins <sergio.martins@kdab.com>
* @since 4.11
*/
class AKONADI_CALENDAR_EXPORT CalendarBase : public MultiCalendar
{
    Q_OBJECT
public:
    typedef QSharedPointer<CalendarBase> Ptr;

    /**
      * Constructs a CalendarBase object.
      */
    explicit CalendarBase(QObject *parent = 0);

    /**
      * Destroys the calendar.
      */
    ~CalendarBase();


    /**
      * Returns the Item containing the incidence with uid @p uid or an invalid Item
      * if the incidence isn't found.
      */
    Akonadi::Item item(const KCalCore::Incidence::Ptr &incidence) const;

    /**
      * Returns the Item with @p id or an invalid Item if not found.
      */
    Akonadi::Item item(Akonadi::Item::Id) const;

    /**
     *  Returns the list of items contained in this calendar.
     *  @see incidences()
     */
    Akonadi::Item::List items() const;

    /**
     * Returns the list of items contained in this calendar that belong to the specified collection.
     * @see incidences()
     * // TODO_KDE5: merge with items() overload
     * @since 4.12
     */
    Akonadi::Item::List items(Akonadi::Collection::Id) const;

    /**
      * Returns the item list that corresponds to the @p incidenceList.
      */
    Akonadi::Item::List itemList(const KCalCore::Incidence::List &incidenceList) const;

    using MultiCalendar::childIncidences;

    /**
      * Returns the child incidences of the parent identified by @p parentId.
      * Only the direct childs are returned
      * @param parentId identifier of the parent item
      */
    KCalCore::Incidence::List childIncidences(const Akonadi::Item::Id &parentId) const;

    Akonadi::Item::List childItems(const KCalCore::Incidence::Ptr &incidence) const;

    /**
      * Returns the child items of the parent identified by @p parentId.
      * Only the direct childs are returned
      * @param parentId identifier of the parent item
      */
    Akonadi::Item::List childItems(const Akonadi::Item::Id &parentId) const;

    /**
      * Sets the weak pointer that's associated with this instance.
      * Use this if later on you need to cast sender() into a QSharedPointer
      *
      * @code
      *   QWeakPointer<CalendarBase> weakPtr = qobject_cast<CalendarBase*>( sender() )->weakPointer();
      *   CalendarBase::Ptr calendar( weakPtr.toStrongRef() );
      * @endcode
      *
      * @see weakPointer()
      */
    void setWeakPointer(const QWeakPointer<Akonadi::CalendarBase> &pointer);

    /**
      * Returns the weak pointer set with setWeakPointer().
      * The default is an invalid weak pointer.
      * @see setWeakPointer()
      */
    QWeakPointer<CalendarBase> weakPointer() const;

    /**
      * Adds an incidence to the calendar.
      * It's added to akonadi in the background @see createFinished().
      * @param incidence the incidence to add
      */
    /**reimp*/ bool addIncidence(const KCalCore::Incidence::Ptr &incidence);
    /**reimp*/ bool addIncidence(const QString &calendar, const KCalCore::Incidence::Ptr &incidence);

    /**
      * Deletes an incidence from the calendar.
      * It's removed from akonadi in the background @see deleteFinished().
      * @param incidence the incidence to delete
      */
    /**reimp*/ bool deleteIncidence(const KCalCore::Incidence::Ptr &);

    /**reimp*/ bool deleteIncidenceInstances(const KCalCore::Incidence::Ptr &);

    /**
        Call this to tell the calendar that you're adding a batch of incidences.
        So it doesn't, for example, ask the destination for each incidence.

        @see endBatchAdding()
    */
    /**reimp*/ void startBatchAdding();

    /**
      * Tells the Calendar that you stoped adding a batch of incidences.
      * @see startBatchAdding()
      */
    /**reimp*/ void endBatchAdding();

    /**
      * Returns the IncidenceChanger used by this calendar to make changes in akonadi.
      * Use this if you need the defaults used by CalendarBase.
      */
    Akonadi::IncidenceChanger *incidenceChanger() const;

    /**
      * Modifies an incidence.
      * The incidence with the same uid as @p newIncidence will be updated with the contents of
      * @param newIncidence the incidence to modify
      */
    bool modifyIncidence(const KCalCore::Incidence::Ptr &oldIncidence, const KCalCore::Incidence::Ptr &newIncidence);

Q_SIGNALS:
    /**
      * This signal is emitted when an incidence is created in akonadi through
      * add{Incidence,Event,Todo,Journal}
      * @param success the success of the operation
      * @param errorMessage if @p success is false, contains the error message
      */
    void createFinished(bool success, const QString &errorMessage);

    /**
      * This signal is emitted when an incidence is deleted in akonadi through
      * delete{Incidence,Event,Todo,Journal} or deleteAll{Events,Todos,Journals}
      * @param success the success of the operation
      * @param errorMessage if @p success is false, contains the error message
      */
    void deleteFinished(bool success, const QString &errorMessage);

    /**
      * This signal is emitted when an incidence is modified in akonadi through
      * modifyIncidence().
      * @param success the success of the operation
      * @param errorMessage if @p success is false, contains the error message
      */
    void modifyFinished(bool success, const QString &errorMessage);

protected:
    Q_DECLARE_PRIVATE(CalendarBase)
    QScopedPointer<CalendarBasePrivate> d_ptr;
    CalendarBase(CalendarBasePrivate *const d, QObject *parent);

    friend class Scheduler;
};
}

#endif
